from dataclasses import dataclass
from typing import Callable, List, Union
import numpy as np

import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import tensorflow as tf

from .op import Op
from .param import get_parameter_values, ppp, upp, bpp

__all__ = ['Mean', 'Trend']

# TODO: in `call` methods, call get_parameter_values beforehand?

def get_trend_coefs(beta):
    if isinstance(beta, (list, tuple)):
        return [p for s in beta for p in upp(s)]
    elif isinstance(beta, str):
        return upp(beta)
    else:
        return []

class Mean(Op):
    def __init__(self, fa, autoinputs):
        if 'locs1' not in autoinputs: autoinputs['locs1'] = 'locs1'
        super().__init__(fa, autoinputs)

    def __add__(self, other):
        if isinstance(other, ZeroTrend):
            return self
        elif isinstance(self, ZeroTrend):
            return other
        else:
            return Stack([self]) + other

    def call(self, p, e):
        pass

    def __call__(self, p, e):
        """
        Returns tuple `(mean, covariance)` for locations.
        Return values have correct shapes.
        """
        M = self.call(p, e)
        if M is None: M = 0.
        n1 = tf.shape(e['locs1'])[0]
        M = tf.broadcast_to(M, [n1])
        return M

class Trend(Mean):
    def __init__(self, featurizer, beta='beta'):
        self.featurizer = featurizer
        super().__init__(dict(beta=beta), dict(locs1='locs1'))

    def vars(self):
        return get_trend_coefs(self.fa['beta'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        x = tf.cast(self.featurizer(e['locs1']), tf.float32)
        if isinstance(v['beta'], (tuple, list)):
            v['beta'] = tf.stack(v['beta'])
        return tf.einsum('ab,b->a', x, v['beta']) # [locs1]

class ZeroTrend(Op):
    def __init__(self):
        super().__init__({}, dict(locs1='locs1'))

    def vars(self):
        return []

    def __call__(self, p, e):
        return tf.zeros_like(e['locs1'][:, 0])

class Mix(Mean):
    def __init__(self, inputs, weights=None):
        fa = {}
        if weights is not None: fa['weights'] = weights
        super().__init__(fa, dict(inputs=inputs, cats1='cats1'))

    def vars(self):
        if 'weights' in self.fa:
            return [p for row in self.fa['weights']
                      for p in get_trend_coefs(row)]
        else:
            return []

    def call(self, p, e):
        M = tf.stack(e['inputs'], axis=-1) # [locs, numinputs].

        # Transform M with weights, if given.
        v = get_parameter_values(self.fa, p)
        if 'weights' in v:
            weights = []
            for row in v['weights']:
                if isinstance(row, (tuple, list)):
                    row = tf.stack(row)
                    weights.append(row)
            weights = tf.stack(weights)
            M = tf.einsum('lh,sh->ls', M, weights)

        return tf.gather(M, e['cats1'], batch_dims=1) # [locs]

class Stack(Mean):
    def __init__(self, parts: List[Mean]):
        self.parts = parts
        super().__init__({}, dict(locs1='locs1', locs2='locs2', parts=parts))

    def vars(self):
        return [p for part in self.parts for p in part.vars()]

    def __add__(self, other):
        if isinstance(other, Mean):
            return Stack(self.parts + [other])

    def call(self, p, e):
        return tf.reduce_sum(e['parts'], axis=0)

    def report(self, p):
        return ' '.join(part.report(p) for part in self.parts)

    def reg(self, p):
        return tf.reduce_sum([part.reg(p) for part in self.parts], axis=0)



from dataclasses import dataclass, replace
from typing import Dict
import tensorflow as tf
import numpy as np
from scipy.special import logit
from .op import SingletonTraceType

@dataclass
class Bound:
    lo: float
    hi: float

    def bounding(self):
        return \
            ('u' if self.lo == float('-inf') else 'b') + \
            ('u' if self.hi == float('inf') else 'b')

    def get_underlying_parameter(self, v, name=None):
        b = self.bounding()
        if b == 'bb':
            init = logit((v - self.lo) / (self.hi - self.lo))
        elif b == 'bu':
            init = np.log(v - self.lo)
        elif b == 'ub':
            init = -np.log(self.hi - v)
        else:
            init = v
        return tf.Variable(init, name=name, dtype=tf.float32)

    def get_surface_parameter(self, v):
        b = self.bounding()
        if b == 'bb':
            v = tf.math.sigmoid(v) * (self.hi - self.lo) + self.lo
        elif b == 'bu':
            v = tf.exp(v) + self.lo
        elif b == 'ub':
            v = self.hi - tf.exp(-v)
        else:
            v = v + tf.constant(0.)
        return v

@dataclass
class ParameterSpace:
    bounds: Dict[str, Bound]
    
    def get_underlying(self, parameters):
        up = {}
        for name, v in parameters.items():
            up[name] = self.bounds[name].get_underlying_parameter(v, name)
        return up

    def get_surface(self, parameters, numpy=False):
        sp = {}
        for name, v in parameters.items():
            v = self.bounds[name].get_surface_parameter(v)
            if numpy: v = v.numpy()
            sp[name] = v
        return sp

    def __tf_tracing_type__(self, context):
        return SingletonTraceType(type(self))

@dataclass(frozen=True)
class PaperParameter:
    name: str
    lo: float
    hi: float

def get_parameter_values(
    blob: object,
    p: Dict[str, object]
):
    """
    For each string encountered in the nested blob,
    look it up in `p` and replace it with the lookup result.
    """
    if isinstance(blob, dict):
        return {k: get_parameter_values(a, p) for k, a in blob.items()}
    elif isinstance(blob, (list, tuple)):
        return [get_parameter_values(a, p) for a in blob]
    elif isinstance(blob, str):
        if blob not in p:
            raise ValueError('Parameter `%s` not found' % blob)
        return p[blob]
    elif blob is None:
        return None
    else:
        return blob

def ppp(name):
    """Positive paper parameter (maybe)."""
    if isinstance(name, str):
        return [PaperParameter(name, 0., float('inf'))]
    else:
        return []

def upp(name):
    """Unbounded paper parameter (maybe)."""
    if isinstance(name, str):
        return [PaperParameter(name, float('-inf'), float('inf'))]
    else:
        return []

def bpp(name, lo, hi):
    """Bounded paper parameter (maybe)."""
    if isinstance(name, str):
        return [PaperParameter(name, lo, hi)]
    else:
        return []

def ppp_list(beta):
    if isinstance(beta, (list, tuple)):
        return [p for s in beta for p in ppp(s)]
    elif isinstance(beta, str):
        return ppp(beta)
    else:
        return []



import numpy as np
from geostat import GP, Model, Featurizer, NormalizingFeaturizer, Mix, Trend
import geostat.kernel as krn
import numpy as np
import tensorflow as tf
from argparse import Namespace


def test_delta():
    # Create 100 random locations in a square centered on the origin.
    np.random.seed(123)
    tf.random.set_seed(123)

    counts = np.array([100, 150, 200, 200]) * 3
    locs1 = np.random.uniform(-1., 1., [sum(counts) // 3, 4]) * [10., 10., 1., 1.]
    locs1 = locs1[:, np.newaxis, :] + np.pad(np.linspace(0.01, 0.1, 3)[:, np.newaxis], ((0, 0), (2, 1)))
    locs1 = locs1.reshape([-1, 4])

    cats1 = [x for i, count in enumerate(counts) for x in [i] * count]

    # Initialize featurizer of location for trends.
    def trend_terms(x, y, z, t): return z, z*z, z*z*z
    featurizer = NormalizingFeaturizer(trend_terms, locs1)

    i_u = GP(0, krn.TrendPrior(featurizer, alpha='au')
                 + krn.SquaredExponential(sill='su1', range='ru1', scale=[1., 1., 'zu', 'tu']))
    i_p = GP(0, krn.TrendPrior(featurizer, alpha='ap')
                 + krn.SquaredExponential(sill='sp1', range='rp1', scale=[1., 1., 'zp', 0.  ]))
    i_t = GP(0, krn.TrendPrior(featurizer, alpha='at')
                 + krn.SquaredExponential(sill='st1', range='rt1', scale=[1., 1., 'zt', 'tt']))

    feat_r = NormalizingFeaturizer(lambda x, y, z, t: (), locs1)

    def neg_transformed_natural_gradient(x, y, z, t):
        degF = 62.2 + 57.1 * (0.15 - z)
        return -(tf.math.log(degF + 6.77) - tf.math.log(75 + 6.77))

    o_u = GP(0, krn.Noise(nugget='nu') + krn.Delta(dsill='wu', axes=[0, 1]))
    o_p = GP(0, krn.Noise(nugget='np') + krn.Delta(dsill='wp', axes=[0, 1]))
    o_t = GP(0, krn.Noise(nugget='nt') + krn.Delta(dsill='wt', axes=[0, 1]))
    o_r = GP(Trend(Featurizer(neg_transformed_natural_gradient), beta=[1.]),
             krn.Noise(nugget='nr') + krn.Delta(dsill='wr', axes=[0, 1])
             + krn.TrendPrior(feat_r, alpha='ar'))

    gp = Mix([i_u, i_p, i_t], [[1., 0., 0.], [0., 1., 0.], [0., 0., 1.], ['cu', 'cp', 0.]]) \
       + Mix([o_u, o_p, o_t, o_r])
        
    p_init = {'au': 1.,
     'su1': 0.01,
     'ru1': 0.5,
     'cu': -0.5,
     'wu': 0.01,
     'nu': 0.02,
     'ap': 1.,
     'sp1': 0.04,
     'rp1': 0.4,
     'cp': -0.3,
     'wp': 0.002,
     'np': 0.001,
     'at': 0.1,
     'st1': 0.02,
     'rt1': 1.,
     'wt': 0.01,
     'nt': 0.01,
     'ar': 1.,
     'wr': 0.03,
     'nr': 0.001,
     'zu': 10.,
     'zp': 15.,
     'zt': 20.,
     'tu': 10.,
     'tt': 10.}

    def report(p):
        p = {k: (v.numpy() if hasattr(v, 'numpy') else v) for k, v in p.items()}
        p = Namespace(**p)
        print(f'[iter {p.iter:5d}] [ll {p.ll:.3f}] [reg {p.reg:.3f}] [time {p.time:.1f}]')
        print(f'   {"alp":>6s} {" zs":>6s} {" ts":>6s} {" sil":>6s} {" rng":>6s} {"  c":>6s} {" wi":>6s} {"nug":>6s}')
        print(f'u: {p.au:6.3f} {p.zu:6.3f} {p.tu:6.3f} {p.su1:6.3f} {p.ru1:6.3f} {p.cu:6.3f} {p.wu:6.3f} {p.nu:6.3f}')
        print(f'p: {p.ap:6.3f} {p.zp:6.3f} {"   ":>6s} {p.sp1:6.3f} {p.rp1:6.3f} {p.cp:6.3f} {p.wp:6.3f} {p.np:6.3f}')
        print(f't: {p.at:6.3f} {p.zt:6.3f} {p.tt:6.3f} {p.st1:6.3f} {p.rt1:6.3f} {-1. :6.3f} {p.wt:6.3f} {p.nt:6.3f}')
        print(f'r: {p.ar:6.3f} {"    ":6s} {"    ":6s} {"     ":6s} {"     ":6s} {"    ":6s} {p.wr:6.3f} {p.nr:6.3f}')

    model1 = Model(
        gp,
        parameters = p_init,
        report = report,
        verbose = True)

    # Generate data.
    vals1 = model1.generate(locs1, cats1).vals

    model2 = Model(
        gp,
        parameters = {k: 2*v if 'g' not in k else v for k, v in p_init.items()},
        report = report,
        verbose=True).fit(locs1, vals1, cats1, reg=1., iters=500)



from argparse import Namespace
import numpy as np
import tensorflow as tf
from geostat import Featurizer, GP, Model, NormalizingFeaturizer, Mix, Trend
import geostat.kernel as krn

def test_multigp():

    np.random.seed(12)
    tf.random.set_seed(12)

    # Create random locations in a square centered on the origin.
    N = 600
    locs1 = np.random.normal(size=[3*N, 2])
    cats1 = [0] * N + [1] * N + [2] * N

    # Initialize featurizer of location for trends.
    def trend_terms(x, y): return x, y, x*y
    featurizer = NormalizingFeaturizer(trend_terms, locs1)
    in1 = GP(0, krn.TrendPrior(featurizer, alpha='a1') + krn.SquaredExponential(sill='s1', range='r1'))
    in2 = GP(0, krn.TrendPrior(featurizer, alpha='a2') + krn.SquaredExponential(sill='s2', range='r2'))

    f2 = Featurizer(lambda x, y: (1, x + y*y))
    out1 = GP(Trend(f2, beta=[0., 0.]), krn.Noise(nugget='n1'))
    out2 = GP(Trend(f2, beta=['c2', 0.]), krn.Noise(nugget='n2'))
    out3 = GP(Trend(f2, beta=[0., 1.]), krn.Noise(nugget='n3'))

    gp = Mix([in1, in2], [[1., 0.], [0., 1.], ['k1', 'k2']]) + Mix([out1, out2, out3])

    def report(p, prefix=''):
        p = {k: (v.numpy() if hasattr(v, 'numpy') else v) for k, v in p.items()}
        p = Namespace(**p)

        if 'iter' in p: print(f'[iter {p.iter:5d}] [ll {p.ll:.3f}] [reg {p.reg:.3f}] [time {p.time:.1f}]')

        print(f"""
            a1={p.a1:.2f}, s1={p.s1:.2f}, r1={p.r1:.2f}, k1={p.k1:.2f},
            a2={p.a2:.2f}, s2={p.s2:.2f}, r2={p.r2:.2f}, k2={p.k2:.2f}, c2={p.c2:.2f},
            n1={p.n1:.2f}, n2={p.n2:.2f}, n3={p.n3:.2f}"""[1:])

    # Generating GP.
    vals1 = Model(
        gp,
        parameters = dict(
            a1=1., s1=1., r1=0.5, k1=2.,
            a2=1., s2=1., r2=0.5, k2=3., c2=1.,
            n1=0.1, n2=0.2, n3=0.3),
        verbose=True).generate(locs1, cats1).vals

    # Fit GP.
    model = Model(
        gp,
        parameters = dict(
            a1=1., s1=1., r1=1., k1=0.,
            a2=1., s2=1., r2=1., k2=0., c2=0.,
            n1=0.1, n2=0.1, n3=0.1),
        report=report,
        verbose=True).fit(locs1, vals1, cats1, iters=5000)

    # Interpolate using GP.
    N = 20
    xx, yy = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy], axis=-1)
    for cats in range(3):
        mean, var = model.predict(locs2, cats * np.ones_like(locs2[..., 0], dtype=np.int32))

    assert np.allclose(
        [model.parameters[p] for p in 's1 r1 k1 s2 r2 k2 c2 n1 n2 n3'.split()],
        [1., 0.5, 2., 1., 0.5, 3., 1., 0.1, 0.2, 0.3],
        rtol=0.5)



import numpy as np
import tensorflow as tf
from geostat import Featurizer, GP, Model, NormalizingFeaturizer, Trend
import geostat.kernel as krn

def test_pairwise_prediction():
    np.random.seed(2)
    tf.random.set_seed(2)

    # Create random locations in a square centered on the origin.
    locs1 = np.random.normal(size=[1000, 2])

    # Initialize featurizer of location for trends.
    def trend_terms(x, y): return x, y, x*y
    featurizer = NormalizingFeaturizer(trend_terms, locs1)
    kernel = krn.TrendPrior(featurizer) + krn.SquaredExponential(sill=1.) + krn.Noise()

    # Generate data.
    vals1 = Model(
        GP(0, kernel),
        parameters = dict(alpha=1., range=0.33, nugget=1.),
        verbose=True).generate(locs1).vals

    # Fit GP.
    model = Model(
        GP(0, kernel),
        parameters = dict(alpha=2., range=1., nugget=0.5),
        verbose=True).fit(locs1, vals1, iters=100, step_size=1e-1)

    assert np.allclose(
        [model.parameters[p] for p in ['range', 'nugget']],
        [0.33, 1.],
        rtol=0.3)

    # Interpolate using GP.
    N = 40
    xx, yy = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy], axis=-1).reshape([-1, 2, 2])

    mean, var = model.predict(locs2, pair=True)



from dataclasses import dataclass
from typing import Callable, List, Union
import numpy as np

# Tensorflow is extraordinarily noisy. Catch warnings during import.
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import tensorflow as tf
    from tensorflow.linalg import LinearOperatorFullMatrix as LOFullMatrix
    from tensorflow.linalg import LinearOperatorBlockDiag as LOBlockDiag

from .op import Op
from .metric import Euclidean, PerAxisDist2, ed
from .param import get_parameter_values, ppp, upp, bpp, ppp_list
from .mean import get_trend_coefs

__all__ = ['Kernel']

def block_diag(blocks):
    """Return a dense block-diagonal matrix."""
    return LOBlockDiag([LOFullMatrix(b) for b in blocks]).to_dense()

class Kernel(Op):
    def __init__(self, fa, autoinputs):
        if 'offset' not in autoinputs: autoinputs['offset'] = 'offset'
        if 'locs1' not in autoinputs: autoinputs['locs1'] = 'locs1'
        if 'locs2' not in autoinputs: autoinputs['locs2'] = 'locs2'
        super().__init__(fa, autoinputs)

    def __add__(self, other):
        if other is None:
            return self
        else:
            return Stack([self]) + other

    def __mul__(self, other):
        return Product([self]) * other

    def call(self, p, e):
        """
        Returns tuple `(mean, covariance)` for locations.
        Return values may be unbroadcasted.
        """
        pass

    def __call__(self, p, e):
        """
        Returns tuple `(mean, covariance)` for locations.
        Return values have correct shapes.
        """
        C = self.call(p, e)
        if C is None: C = 0.
        n1 = tf.shape(e['locs1'])[0]
        n2 = tf.shape(e['locs2'])[0]
        C = tf.broadcast_to(C, [n1, n2])
        return C

    def report(self, p):
        string = ', '.join('%s %4.2f' % (v.name, p[v.name]) for v in self.vars())
        return '[' + string + ']'

    def reg(self, p):
        pass

class TrendPrior(Kernel):
    def __init__(self, featurizer, alpha='alpha'):
        fa = dict(alpha=alpha)
        self.featurizer = featurizer
        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['alpha'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        F1 = tf.cast(self.featurizer(e['locs1']), tf.float32)
        F2 = tf.cast(self.featurizer(e['locs2']), tf.float32)
        return v['alpha'] * tf.einsum('ba,ca->bc', F1, F2)

    def reg(self, p):
        return 0.

def scale_to_metric(scale, metric):
    assert scale is None or metric is None
    if metric is None:
        if scale is None:
            metric = 'euclidean'
        else:
            metric = Euclidean(scale)
    return metric

class SquaredExponential(Kernel):
    def __init__(self, sill='sill', range='range', scale=None, metric=None):
        fa = dict(sill=sill, range=range)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp(self.fa['sill']) + ppp(self.fa['range'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        return v['sill'] * tf.exp(-0.5 * e['d2'] / tf.square(v['range']))

    def reg(self, p):
        v = get_parameter_values(self.fa, p)
        return v['range']

class GammaExponential(Kernel):
    def __init__(self, range='range', sill='sill', gamma='gamma', scale=None, metric=None):
        fa = dict(sill=sill, range=range, gamma=gamma, scale=scale)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp(self.fa['sill']) + ppp(self.fa['range']) + bpp(self.fa['gamma'], 0., 2.)

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        return v['sill'] * gamma_exp(e['d2'] / tf.square(v['range']), v['gamma'])

    def reg(self, p):
        v = get_parameter_values(self.fa, p)
        return v['range']

@tf.custom_gradient
def ramp(x):
    ax = tf.abs(x)
    def grad(upstream):
        return upstream * tf.where(ax < 1., -tf.sign(x), 0.)
    return tf.maximum(0., 1. - ax), grad

class Ramp(Kernel):
    def __init__(self, range='range', sill='sill', scale=None, metric=None):
        fa = dict(sill=sill, range=range, scale=scale)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp(self.fa['sill']) + ppp(self.fa['range'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        # return v['sill'] * tf.clip_by_value(1. - tf.sqrt(e['d2']) / v['range'], 0., 1.)
        return v['sill'] * ramp(tf.sqrt(e['d2']) / v['range'])

    def reg(self, p):
        v = get_parameter_values(self.fa, p)
        return v['range']

# @tf.custom_gradient
# def rampstack(x, sills, ranges):
#     """
#     `x` has arbitrary shape [...].
#     `sills` and `ranges` both have shape [K].
#     """
#     ax = ed(tf.abs(x)) # [..., 1]
#     y = sills * tf.maximum(0., 1. - ax / ranges) # [..., K]
#     def grad(upstream):
#         ax = ed(tf.abs(x)) # [..., 1]
#         y = sills * tf.maximum(0., 1. - ax / ranges) # [..., K]
#         K = tf.shape(sills)[0]
#         grad_x = upstream * tf.reduce_sum(tf.where(ax < ranges, -tf.sign(ed(x)), 0.), -1) # [...]
#         grad_sills = tf.reduce_sum(tf.reshape(ed(upstream) * y, [-1, K]), 0) # [K}
#         grad_ranges = tf.where(ax < ranges, sills * ax / tf.square(ranges), 0.) # [..., K}
#         grad_ranges = tf.reduce_sum(tf.reshape(ed(upstream) * grad_ranges, [-1, K]), 0) # [K]
#         return grad_x, grad_sills, grad_ranges
#     return tf.reduce_sum(y, -1), grad

@tf.custom_gradient
def rampstack(x, sills, ranges):
    """
    `x` has arbitrary shape [...], but must be non-negative.
    `sills` and `ranges` both have shape [K].
    """
    ax = ed(tf.abs(x)) # [..., 1]
    y = sills * tf.maximum(0., 1. - ax / ranges) # [..., K]
    def grad(upstream):
        ax = ed(tf.abs(x)) # [..., 1]
        y = sills * tf.maximum(0., 1. - ax / ranges) # [..., K]
        K = tf.shape(sills)[0]
        small = ax < ranges
        grad_x = upstream * tf.reduce_sum(tf.where(small, -tf.sign(ed(x)) * (sills / ranges), 0.), -1) # [...]
        grad_sills = tf.einsum('ak,a->k', tf.reshape(y, [-1, K]), tf.reshape(upstream, [-1]))
        grad_ranges = tf.where(small, ax * (sills / tf.square(ranges)), 0.) # [..., K}
        grad_ranges = tf.einsum('ak,a->k', tf.reshape(grad_ranges, [-1, K]), tf.reshape(upstream, [-1]))
        return grad_x, grad_sills, grad_ranges
    return tf.reduce_sum(y, -1), grad

class RampStack(Kernel):
    def __init__(self, range='range', sill='sill', scale=None, metric=None):
        fa = dict(sill=sill, range=range, scale=scale)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp_list(self.fa['sill']) + ppp_list(self.fa['range'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        if isinstance(v['sill'], (tuple, list)):
            v['sill'] = tf.stack(v['sill'])
        if isinstance(v['range'], (tuple, list)):
            v['range'] = tf.stack(v['range'])

        return rampstack(tf.sqrt(e['d2']), v['sill'], v['range'])

    def reg(self, p):
        v = get_parameter_values(self.fa, p)
        return v['range']

@tf.recompute_grad
def smooth_convex(x, sills, ranges):
    """
    `x` has arbitrary shape [...], but must be non-negative.
    `sills` and `ranges` both have shape [K].
    """
    r2 = ranges
    r1 = tf.pad(ranges[:-1], [[1, 0]])
    ex = ed(x)
    ax = tf.abs(ex)
    rx = ax / r2 - 1.

    c1 = 2. / (r1 + r2)
    c2 = 1. / (1. - tf.square(r1/r2))

    # i1 = tf.cast(ax <= r1, tf.float32) # Indicates x <= r1.
    # i2 = tf.cast(ax <= r2, tf.float32) * (1. - i1) # Indicates r1 < x <= r2.
    # v = i1 * (1. - c1 * ax) + i2 * c2 * tf.square(rx)

    v = tf.where(ax <= r1, 1. - c1 * ax, c2 * tf.square(rx))
    v = tf.where(ax <= r2, v, 0.)

    y = tf.einsum('...k,k->...', v, sills)
    return y

@tf.custom_gradient
def smooth_convex_grad(x, sills, ranges):
    """
    `x` has arbitrary shape [...], but must be non-negative.
    `sills` and `ranges` both have shape [K].
    """
    r2 = ranges
    r1 = tf.pad(ranges[:-1], [[1, 0]])
    ex = ed(x)
    ax = tf.abs(ex)
    rx = ax / r2 - 1.

    c1 = 2. / (r1 + r2)
    c2 = 1. / (1. - tf.square(r1/r2))

    v = tf.where(ax <= r1, 1. - c1 * ax, c2 * tf.square(rx))
    v = tf.where(ax <= r2, v, 0.)

    y = tf.einsum('...k,k->...', v, sills)

    def grad(upstream):
        r2 = ranges
        r1 = tf.pad(ranges[:-1], [[1, 0]])
        ex = ed(x)
        ax = tf.abs(ex)
        rx = ax / r2 - 1.
        i1 = tf.cast(ax <= r1, tf.float32) # Indicates x <= r1.
        i2 = tf.cast(ax <= r2, tf.float32) * (1. - i1) # Indicates r1 < x <= r2.

        c1 = 2. / (r1 + r2)
        c2 = 1. / (1. - tf.square(r1 / r2))
        c3 = 1. / (r2 - tf.square(r1) / r2)

        v = i1 * (1. - c1 * ax) + i2 * c2 * tf.square(rx)

        sx = tf.sign(ex)

        K = tf.shape(sills)[0]
        gx = sx * sills * (i1 * -c1 + i2 * rx * (2 * c3))
        grad_x = upstream * tf.reduce_sum(gx, -1) # [...]

        grad_sills = tf.einsum('ak,a->k', tf.reshape(v, [-1, K]), tf.reshape(upstream, [-1]))

        u = 2 / tf.square(r1 + r2) * ax * i1
        yr1 = u + i2 * tf.square(rx * c3) * 2 * r1
        yr2 = u - 2 * i2 * (rx * c3 + tf.square(rx * c2) / r2)
        yr1 = sills * tf.reshape(yr1, [-1, K])
        yr2 = sills * tf.reshape(yr2, [-1, K])
        yr = tf.pad(yr1[:, 1:], [[0, 0], [0, 1]]) + yr2
        grad_ranges = tf.einsum('ak,a->k', yr, tf.reshape(upstream, [-1]))

        return grad_x, grad_sills, grad_ranges

    return y, grad

class SmoothConvex(Kernel):
    def __init__(self, range='range', sill='sill', scale=None, metric=None):
        fa = dict(sill=sill, range=range, scale=scale)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp_list(self.fa['sill']) + ppp_list(self.fa['range'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        if isinstance(v['sill'], (tuple, list)):
            v['sill'] = tf.stack(v['sill'])
        if isinstance(v['range'], (tuple, list)):
            v['range'] = tf.stack(v['range'])

        return smooth_convex(tf.sqrt(e['d2']), v['sill'], v['range'])

    def reg(self, p):
        v = get_parameter_values(self.fa, p)
        return v['range']

@tf.recompute_grad
def quadstack(x, sills, ranges):
    """
    `x` has arbitrary shape [...], but must be non-negative.
    `sills` and `ranges` both have shape [K].
    """
    ex = ed(x)
    ax = tf.maximum(0., 1. - tf.abs(ex) / ranges) # [..., 1]
    y = sills * tf.square(ax) # [..., K]
    return tf.reduce_sum(y, -1)

class QuadStack(Kernel):
    def __init__(self, range='range', sill='sill', scale=None, metric=None):
        fa = dict(sill=sill, range=range, scale=scale)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp_list(self.fa['sill']) + ppp_list(self.fa['range'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        if isinstance(v['sill'], (tuple, list)):
            v['sill'] = tf.stack(v['sill'])
        if isinstance(v['range'], (tuple, list)):
            v['range'] = tf.stack(v['range'])

        return quadstack(tf.sqrt(e['d2']), v['sill'], v['range'])

    def reg(self, p):
        v = get_parameter_values(self.fa, p)
        return v['range']

class Wiener(Kernel):
    def __init__(self, axis, start):

        self.axis = axis
        self.start = start

        # Include the element of scale corresponding to the axis of
        # integration as an explicit formal argument.
        fa = dict()

        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return []

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        x1 = e['locs1'][..., self.axis]
        x2 = e['locs2'][..., self.axis]
        k = tf.maximum(0., tf.minimum(ed(x1, 1), ed(x2, 0)) - self.start)
        return k

    def reg(self, p):
        return 0.

class IntSquaredExponential(Kernel):
    def __init__(self, axis, start, range='range'):

        self.axis = axis
        self.start = start

        # Include the element of scale corresponding to the axis of
        # integration as an explicit formal argument.
        fa = dict(range=range)

        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['range'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        x1 = tf.pad(e['locs1'][..., self.axis] - self.start, [[1, 0]])
        x2 = tf.pad(e['locs2'][..., self.axis] - self.start, [[1, 0]])

        r = v['range']
        sdiff = (ed(x1, 1) - ed(x2, 0)) / (r * np.sqrt(2.))
        k = -tf.square(r) * (np.sqrt(np.pi) * sdiff * tf.math.erf(sdiff) + tf.exp(-tf.square(sdiff)))
        k -= k[0:1, :]
        k -= k[:, 0:1]
        k = k[1:, 1:]
        k = tf.maximum(0., k)

        return k

    def reg(self, p):
        return 0.

class IntExponential(Kernel):
    def __init__(self, axis, start, range='range'):

        self.axis = axis
        self.start = start

        # Include the element of scale corresponding to the axis of
        # integration as an explicit formal argument.
        fa = dict(range=range)

        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['range'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        x1 = tf.pad(e['locs1'][..., self.axis] - self.start, [[1, 0]])
        x2 = tf.pad(e['locs2'][..., self.axis] - self.start, [[1, 0]])

        r = v['range']
        sdiff = tf.abs(ed(x1, 1) - ed(x2, 0)) / r
        k = -tf.square(r) * (sdiff + tf.exp(-sdiff))
        k -= k[0:1, :]
        k -= k[:, 0:1]
        k = k[1:, 1:]
        k = tf.maximum(0., k)

        return k

    def reg(self, p):
        return 0.

class Noise(Kernel):
    def __init__(self, nugget='nugget'):
        fa = dict(nugget=nugget)
        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['nugget'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)

        indices1 = tf.range(tf.shape(e['locs1'])[0])
        indices2 = tf.range(tf.shape(e['locs2'])[0]) + e['offset']
        C = tf.where(tf.equal(tf.expand_dims(indices1, -1), indices2), v['nugget'], 0.)
        return C

    def reg(self, p):
        return 0.

class Delta(Kernel):
    def __init__(self, dsill='dsill', axes=None):
        fa = dict(dsill=dsill)
        self.axes = axes
        super().__init__(fa, dict(pa_d2='per_axis_dist2'))

    def vars(self):
        return ppp(self.fa['dsill'])

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)

        if self.axes is not None:
            n = tf.shape(e['pa_d2'])[-1]
            mask = tf.math.bincount(self.axes, minlength=n, maxlength=n, dtype=tf.float32)
            d2 = tf.einsum('abc,c->ab', e['pa_d2'], mask)
        else:
            d2 = tf.reduce_sum(e['pa_d2'], axis=-1)

        return v['dsill'] * tf.cast(tf.equal(d2, 0.), tf.float32)

    def reg(self, p):
        return 0.

class Mix(Kernel):
    def __init__(self, inputs, weights=None):
        self.inputs = inputs
        fa = {}
        ai = dict(cats1='cats1', cats2='cats2')

        # Special case if weights is not given.
        if weights is not None:
            fa['weights'] = weights
            ai['inputs'] = inputs

        super().__init__(fa, ai)

    def gather_vars(self, cache=None):
        """Make a special version of gather_vars because
           we want to gather variables from `inputs`
           even when it's not in autoinputs"""
        vv = super().gather_vars(cache)
        for iput in self.inputs:
            cache[id(self)] |= iput.gather_vars(cache)
        return cache[id(self)]

    def vars(self):
        if 'weights' in self.fa:
            return [p for row in self.fa['weights']
                      for p in get_trend_coefs(row)]
        else:
            return []

    def call(self, p, e):
        v = get_parameter_values(self.fa, p)
        if 'weights' in v:
            weights = []
            for row in v['weights']:
                if isinstance(row, (tuple, list)):
                    row = tf.stack(row)
                    weights.append(row)
            weights = tf.stack(weights)
            C = tf.stack(e['inputs'], axis=-1) # [locs, locs, numinputs].
            Aaug1 = tf.gather(weights, e['cats1']) # [locs, numinputs].
            Aaug2 = tf.gather(weights, e['cats2']) # [locs, numinputs].
            outer = tf.einsum('ac,bc->abc', Aaug1, Aaug2) # [locs, locs, numinputs].
            C = tf.einsum('abc,abc->ab', C, outer) # [locs, locs].
            return C
        else:
            # When weights is not given, exploit the fact that we don't have
            # to compute every element in component covariance matrices.
            N = len(self.inputs)
            catcounts1 = tf.math.bincount(e['cats1'], minlength=N, maxlength=N)
            catcounts2 = tf.math.bincount(e['cats2'], minlength=N, maxlength=N)
            catindices1 = tf.math.cumsum(catcounts1, exclusive=True)
            catindices2 = tf.math.cumsum(catcounts2, exclusive=True)
            catdiffs = tf.unstack(catindices2 - catindices1, num=N)
            locsegs1 = tf.split(e['locs1'], catcounts1, num=N)
            locsegs2 = tf.split(e['locs2'], catcounts2, num=N)

            CC = [] # Observation noise submatrices.
            for sublocs1, sublocs2, catdiff, iput in zip(locsegs1, locsegs2, catdiffs, self.inputs):
                cache = dict(
                    offset = e['offset'] + catdiff,
                    locs1 = sublocs1,
                    locs2 = sublocs2)
                cache['per_axis_dist2'] = PerAxisDist2().run(cache, p)
                cache['euclidean'] = Euclidean().run(cache, p)
                Csub = iput.run(cache, p)
                CC.append(Csub)

            return block_diag(CC)

    def reg(self, sp):
        return tf.reduce_sum([iput.reg(sp) for iput in self.inputs])

class Stack(Kernel):
    def __init__(self, parts: List[Kernel]):
        self.parts = parts
        super().__init__({}, dict(locs1='locs1', locs2='locs2', parts=parts))

    def vars(self):
        return [p for part in self.parts for p in part.vars()]

    def __add__(self, other):
        if isinstance(other, Kernel):
            return Stack(self.parts + [other])
    
    def call(self, p, e):
        return tf.reduce_sum(e['parts'], axis=0)

    def report(self, p):
        return ' '.join(part.report(p) for part in self.parts)

    def reg(self, p):
        return tf.reduce_sum([part.reg(p) for part in self.parts], axis=0)

class Product(Kernel):
    def __init__(self, parts: List[Kernel]):
        self.parts = parts
        super().__init__({}, dict(locs1='locs1', locs2='locs2', parts=parts))

    def vars(self):
        return [p for part in self.parts for p in part.vars()]

    def __mul__(self, other):
        if isinstance(other, Kernel):
            return Product(self.parts + [other])
    
    def call(self, p, e):
        return tf.reduce_prod(e['parts'], axis=0)

    def report(self, p):
        return ' '.join(part.report(p) for part in self.parts)

    def reg(self, p):
        return tf.reduce_sum([part.reg(p) for part in self.parts], axis=0)

# Gamma exponential covariance function.
@tf.custom_gradient
def safepow(x, a):
    y = tf.pow(x, a)
    def grad(dy):
        dx = tf.where(x <= 0.0, tf.zeros_like(x), dy * tf.pow(x, a-1))
        dx = unbroadcast(dx, x.shape)
        da = tf.where(x <= 0.0, tf.zeros_like(a), dy * y * tf.math.log(x))
        da = unbroadcast(da, a.shape)
        return dx, da
    return y, grad

def unbroadcast(x, shape):
    xrank = len(x.shape)
    rank = len(shape)
    excess_rank = tf.maximum(0, xrank - rank)
    x = tf.reduce_sum(x, axis=tf.range(excess_rank))
    axes_that_are_one = tf.where(tf.equal(shape, 1))[:, 0]
    x = tf.reduce_sum(x, axis=axes_that_are_one, keepdims=True)
    return x

def gamma_exp(d2, gamma):
    return tf.exp(-safepow(tf.maximum(d2, 0.0), 0.5 * gamma))

class Observation(Op):

    def __init__(self,
        coefs: List,
        noise: Kernel
    ):
        self.coefs = coefs
        self.noise = noise
        super().__init__({}, self.noise)

    def vars(self):
        vv = [p for c in self.coefs for p in upp(c)]
        vv += self.noise.vars()
        return vv

    def __call__(self, p, e):
        """
        Dummy.
        """
        return 0.



import pandas as pd
from geostat import Krige
import pyproj

def featurization(x1):
    return x1[:, 0], x1[:, 1]

def project(x):
    return 

def test_krige():
    df = pd.DataFrame().assign(
        lat = [35.0, 35.0, 36.0, 36.0, 37.0, 37.0, 38.0, 38.0, 39.0, 39.0],
        lon = [-120.0, -121.0, -120.0, -121.0, -120.0, -121.0, -120.0, -121.0, -120.0, -121.0],
        u = [5., 5., 5., 6., 5., 7., 6., 6., 6., 7.])

    x1 = df[['lon', 'lat']].to_numpy()

    u1 = df[['u']].to_numpy()

    # Projection
    p1 = pyproj.Proj(proj='latlong', datum='NAD83')
    p2 = pyproj.Proj('EPSG:3310')
    project = pyproj.Transformer.from_proj(p1, p2)
    to_km = lambda x, y: (x * 1e-3, y * 1e-3)

    krige = Krige(x1, u1, 2,
            variogram_func='linear',
            projection=lambda x, y: to_km(*project.transform(x, y)),
            featurization=featurization,
            show_plots=False,
            verbose=True)

    df2 = pd.DataFrame().assign(
        lat = [35.5, 35.5, 36.5],
        lon = [-120.0, -121.0, -121.0])

    x2 = df2[['lon', 'lat']].to_numpy()

    u2_mean, u2_var = krige.predict(x2)
    
    print(u2_mean, u2_var)



from .krige import *
from .mesh import *
from .model import *
from .mean import *

__version__ = '0.9.1'

__all__ = []
__all__.extend(mean.__all__)
__all__.extend(mesh.__all__)
__all__.extend(model.__all__)
__all__.extend(kernel.__all__)
__all__.extend(krige.__all__)
